#!/usr/bin/env python
#coding:utf-8
# Name: JAGuaR - Junction Alignment to Genome for Repositioing RNA-seq Reads
# Author:  Yaron Butterfield
# Purpose: Convert junction aligned reads to genome coordinates ('reposition')
# Original Idea and design:  July 9, 2010
# Redesign to work with longer reads : June 13, 2011
# Released for testing: November 15, 2011
# Released for public use: January, 2014
version = "2.1"
versionDate = "June 26, 2013"
info = "Copyright (c) 2013 Canada's Michael Smith Genome Sciences Centre. All rights reserved."

import sys, time, os, string
import unittest
import samIterator, samFlags, convertCigar
from samIterator import align
from optparse import OptionParser
import difflib  
import cProfile
import subprocess
t0 = time.time()
import re

"""
 (M)ax separation recommendations:
 34622 is used for hg19a
 23545 is used for hg18
"""

trim_for_junction = 2
junctionPadLength=6
cigarFile=''
printInfo=0

linesToReport=1000000
genomeInfo={}
selectChromosomes=[]

"""
SAM FORMAT

Col Field Description 

1 QNAME Query template/pair NAME 
2 FLAG bitwise FLAG 
3 RNAME Reference sequence NAME 
4 POS 1-based leftmost POSition/coordinate of clipped sequence 
5 MAPQ MAPping Quality (Phred-scaled) 
6 CIGAR extended CIGAR string 
7 MRNM Mate Reference sequence NaMe (‘=’ if same as RNAME) 
8 MPOS 1-based Mate POSistion 
9 TLEN inferred Template LENgth (insert size) 
10 SEQ query SEQuence on the same strand as the reference 
11 QUAL query QUALity (ASCII-33 gives the Phred base quality) 
12+ OPT variable OPTional fields in the format TAG:VTYPE:VALUE 

"""

class junction(object):
      """
       Junction reference object
      """
      __slots__ = ['id','genomeRange','sequencelength','exonBoundaries']
      def __init__(self):
            self.__class__.__name__

            self.genomeRange=[]   #needed
            self.sequencelength=0 #needed
            self.exonBoundaries=[] #needed
            self.id=''    # basically count in list 

      def genomeRangeSort(self):
            # sort and convert to integer
            grange = self.genomeRange
            sortedRange = []
            for r in grange:
                  sortedRange.append(int(r))
            sortedRange.sort()
            self.genomeRange = sortedRange


class Transcript(object):
      def __init__(self):
            self.__class__.__name__
            self.id=''
            self.start=''
            self.exon_meta={} # exon rank in this transcript : exon id


class Stats(object):
      def __init__(self):
            self.__class__.__name__
            self.sam_file = ''
            self.new_sam_file = ''
            self.reposMinutes = ''
            self.startTime = ''
            self.unMapped = 0
            self.setUnmapped = 0
            self.totalReads = 0
            self.junctionReads = 0
            self.junctionAligned = 0
            self.junctionRepositionedReads = 0
            self.junctionReadNoN = 0
            self.junctionCoverageTotal = 0 # total number of instances of junctions being spanned
            self.junctionNumberCount = {} # number of times one read covers 1, 2, 3... junctions
            self.junctionUniqueCoverageCount = {} # unique junctions coverage
            self.junctionUniqueCoverageCountTotal=0 # unique junction count
            self.mismatchChromosomePairs = 0 


def getRefGenome(r):
      # read entire reference fasta sequence into memory
      # create sequence object for each chromosome?
      # if this is too slow, perhaps better to read separate chromosome files?

      chrSeq = {}

      x = open(r,'r')
      seq = ''
      c=''
      i=0
      for line in x:

            if ">" in line:
                  if i>0:
                        chrSeq[c]=seq

                  i+=1
                  c = line.split()[0][1:]
                  continue

            seq+=line[:-1].upper()

      return chrSeq


def getChromosomeInfo(c):
      # this be added to the getRefGenome function?  
      # why load up two file?
      genomeInfo = {}
      chrOrder = []

      x = open(c,'r')
      for line in x:
            i = line.split()
            chromosome = i[0]
            chrOrder.append(chromosome)
            if len(selectChromosomes)>0:
                  if chromosome not in selectChromosomes:
                        continue
            chrLength = i[1]
            chrFastaLines = i[2][:-1]
            genomeInfo[chromosome]=int(chrLength)

      return genomeInfo,chrOrder


def get_SQ_info(sqFile):
      SQ_info = []
      x = open(sqFile,'r')
      for line in x:
            SQ_info.append(line)
      return SQ_info


def printStats(outputStub, s):

      summaryFile = outputStub+".Summary.txt"
      summary = open(summaryFile,'w')  

      junctionNo = []
      for noJunctions in s.junctionNumberCount:
            line = "    " + str(noJunctions) + " " + str(s.junctionNumberCount[noJunctions])
            junctionNo.append(line)

      toPrint = [" ",
                 "JAGuaR Summary",
                 "-----------------------------------------------------------------------------------",
                 "Start:  " + s.startTime,
                 "Finish: " + time.ctime(),
                 "File analyzed: " + s.sam_file,
                 "Output file with repositioned reads: "+ s.new_sam_file,
                 "Reference Index Folder: " + s.referenceIndexFolder,
                 "Repositioning time: " + s.reposMinutes + " minutes",
                 "Max Separation: "+str(max_separation),
                 "  ",
                 "Total number of reads: "+ str(s.totalReads),
                 "Total number of reads aligned to junctions: "+str(s.junctionReads),
                 "  ",
                 "Junction reads already unmapped in first alignment: "+str(s.unMapped),
                 "Junction reads set unmapped after repositioning: "+str(s.setUnmapped),
                 "Junction reads mapped and repositioned with large gaps: "+str(s.junctionRepositionedReads),
                 "Junction reads repositioned without large gaps: "+str(s.junctionReadNoN),
                 ]

      if len(junctionNo)==0:
            toPrint.append("Number of times one read covers 1, 2, 3 or more junctions: 0")
      else:
            toPrint.append("Number of times one read covers 1, 2, 3 or more junctions: ")
            toPrint += junctionNo

      toPrint2 = ["Total number of times an exon-exon junction is covered: "+str(s.junctionCoverageTotal),
                  "Total number of exon-exon junctions covered by at least one read: "+str(s.junctionUniqueCoverageCountTotal),
                  "--"
                  ]

      toPrint += toPrint2

      for line in toPrint:
            summary.write(line)
            summary.write('\n')

      return 1


def getJunctionInfo(g,indexFolder):

      junctionIdx = {}


      for chromosome in g:
            count=0

            # The folder containing these files should either be in the options or in a resource file (which contains the path for these files for all 
            #junctionFile = indexFolder+"/"+"chr"+chromosome+"_JN.fa"  
            junctionFile = indexFolder+"/"+chromosome+"_JN.fa"

            j = open(junctionFile,'r')
            exonBoundries = []

            junctionInfo = []

            for line in j:
                  """
	    >31058282,31058354,31060757,31060829|2047-2198
	    """
                  if line[:1]==">":
                        junctionObj = junction()
                        i = line.split("|")

                        boundryRange = i[1].split("-")  # depends on junction fa header format
                        boundryRange[0]=int(boundryRange[0])
                        boundryRange[1]=int(boundryRange[1])
                        b = [boundryRange[0],boundryRange[1]]
                        exonBoundries.append(b)


                        #junctionObj.chromosome = chromosome
                        #junctionObj.id = count
                        junctionObj.exonBoundaries=b
                        junctionObj.genomeRange=i[0][1:].split(',')  # exon start and stop coordinates for all exomes
                        #junctionObj.junctionRefStart = boundryRange[0] # start coordinate in the junction reference
                        #junctionObj.junctionRefStop = boundryRange[1]  # end coordinate in the junction reference
                        #junctionObj.exons = i[0][1:].split(",")  # list of exon names either from Ensembl or somewhere else

                        junctionObj.genomeRangeSort()
                        a=1

                  else:
                        junctionObj.sequencelength += len(line[:-1])
                        #junctionObj.sequence += line[:-1]
                        junctionInfo.append(junctionObj)
                        count+=1

            # create junction index for quick search
            #--------------------------------------------------------------------
            idx={}
            jObjL=[]

            j = 0
            global steps
            steps=8
            # every step you take, every code you break, I'll be watching you...
            for jObj in junctionInfo:
                  jObjL.append(jObj)
                  r = xrange(jObj.exonBoundaries[0],jObj.exonBoundaries[1]+1,steps)
                  for ri in r:
                        idx[ri]=j
                  j+=1
            #---------------------------------------------------------------------  

            jObjL=tuple(jObjL)
            junctionIdx[chromosome]=[idx,jObjL]
      return junctionIdx


def pullOutReadFromSam(sam, reads):
      r = reads.split(",")
      f = open(sam,'r')

      newSam = 'reads.sam'
      newFile = open(newSam,'w')

      select=0
      count=1
      pair=0
      for line in f:
            if len(r)==0:
                  break
            rCount=0
            for qname in r:
                  rCount=1
                  if qname in line:
                        newFile.write(line)
                        #print line[:-1]
                        pair+=1
                        if pair == 2:
                              pair=0
                              r.pop(rCount-1)



def reformat(z,jidx,checksFile):
      #print "-----------------------------"
      minutesElapsed = (time.time()- t0)/60
      minutesElapsed = "%2.0f" % minutesElapsed

      """
     Each exon has a range of coordinates within the junction reference.
     Identify where the read aligned specifically
    """
      seqInfo = z.qname + ", " + str(z.flag)

      chromosome = z.rname
      chrLength = genomeInfo[chromosome]

      alignmentPos = z.pos
      junctionCoordinate = int(alignmentPos) - int(chrLength)


      idx,jObjL= jidx[chromosome]
      jId=''
      jObj=''
      """
     What is the top speed of a Jaguar?
     A Jaguar can run up to 60mph and the Jaguar xk8 can go to 157mph.
     Jaguar repositioning is clocked at just over a minute/million reads.
    """
      stepIt=0
      while stepIt < steps:
            try:
                  jId = idx[junctionCoordinate-stepIt]
                  break
            except KeyError:
                  stepIt+=1

      try:
            jObj = jObjL[jId]
      except:
            # this really shouldn't happen but just in case (<< this does happen)
            z.setUnmapped = 1
            z.cigar = "*"
            return z	

      #print "=>",jObj.exonBoundaries, jObj.genomeRange
      coords = jObj.exonBoundaries

      jC = convertCigar.read_cigar(z.cigar)
      #j1 = [[int(y[0]),y[1]] for y in (re.split('([A-Z])', x)[0:2] for x in jC)]   <-- this is an equivalent alternative to the next three lines
      j1=[]
      for k in jC:
            j1.append([int(k[:-1]),k[-1]])

      # saves softclip info from original cigar	
      if j1[0][1]=='S':
            z.softclip=j1[0][0]
      if j1[-1][1]=='S':
            z.softclipEnd=j1[-1][0]

      """
      a1       a2     b1   b2     c1      c2
       [========]------[====]------[=======]
	    ^ --        ----        -- ^
    """

      junctionStart = int(junctionCoordinate)-int(coords[0])+1

      distanceFromJunctionEnd = (int(coords[1])- int(junctionCoordinate)+1)# from start of read
      z.distanceFromJunctionEnd=distanceFromJunctionEnd

      distanceEndFromJunctionEnd = int(coords[1]) - int(junctionCoordinate + seqLength)+1
      padEnd = seqLength + junctionPadLength # read length + padded sequence (N's)

      if distanceFromJunctionEnd <= 10:

            """
	 We assume that the start of a read aligned to the previous junction reference
	 usually as such (where 'b' is the junction sequence of interest): 

	 -----a-----NNNNNN--------b-------- junction reference
	          ---------------- read

	"""
            z.genomeAligned=1 # should have aligned to genome not junction (?)  Why did it?

            # this read is assigned to the next junction object
            try:
                  jObj = jObjL[jId+1]

            except IndexError:
                  #z.pos = 0
                  z.rname = "*"
                  #z.mapq = "0"
                  z.cigar = "*"
                  z.setUnmapped = 1

                  return z

            junctionStart = 1 

            z.readAlignedToPadStart = distanceFromJunctionEnd   
         
                  

      elif distanceFromJunctionEnd > 10 and distanceFromJunctionEnd < (seqLength+junctionPadLength) and distanceFromJunctionEnd >= seqLength:

            """
	 We assume that the read aligned to the junction sequence such that the end overlaps 
	 with the end of junction padded sequence (where 'a' is the junction sequence of 
	 interest):

	 -------a--------NNNNNN--------b-------- junction reference
		 -----------     read
	"""

            #z.softclipEnd = distanceEndFromJunctionEnd 
            readAlignedToPad = junctionPadLength - distanceEndFromJunctionEnd
            z.readAlignedToPad = readAlignedToPad 

      elif distanceFromJunctionEnd > 10 and distanceFromJunctionEnd < seqLength: 

            """
	 We assume that the read aligned with a portion of the end of the read aligning 
	 into the next junction sequence (where 'a' is the junction sequence of interest).  
	 This should be very rare:

	  ---------a-------NNNNN------b-------
	           -------------------  read
	"""
            #z.softclipEnd = seqLength-distanceFromJunctionEnd
            if printInfo==1:
                  print z.qname, z.rname, z.pos, z.seq  

            if abs(distanceEndFromJunctionEnd) > (10 - junctionPadLength): #throw out reads that have more than 10 bases beyond real junction sequence
                  z.pos = '0'
                  z.mapq = '0'
                  z.rname = '*'
                  z.mrnm = '*'
                  z.cigar = '*'
                  z.setUnmapped = 1
                  return z	    
            else:   # keep reads that have 10 or fewer bases aligned beyond real junction sequence (reads in pad are part of the 10)
                  readAlignedToPad = abs(distanceEndFromJunctionEnd) + junctionPadLength #not sure if this is really necessary, 
                  z.readAlignedToPad = readAlignedToPad 	    

      if printInfo==1: 
            print junctionStart    
      z = reposition(jObj,junctionStart,junctionCoordinate,z,checksFile)

      return z
      

def progressUpdate(lineCount,timeCount):
      minutesElapsed = (time.time()- t0)/60
      minutesElapsed = "%2.0f" % minutesElapsed
      print "Lines processed:",lineCount,minutesElapsed,"minutes elapsed."


def samTagUpdate(z):
      #this function updates the NM and MD and XN and XM tags of reads that have been softclipped.
      #possibly other bwa optional tags (XO XG) will be added later
      
      #save information from pre-repos MD string and NM value
      tag_details = samIterator.getTagDetails(z)
      if 'MD' in tag_details and 'NM' in tag_details:
            orig_MD = tag_details['MD']
            NM = tag_details['NM']
      else:
            return z

      MD_split = re.split("(\d+)", orig_MD)
      MD_split = filter(bool, MD_split) # get rid of empty elements
      MD_n_total = 0
      for element in MD_split:
            if element == "0":
                  continue
            if re.match("[ACTG]", element):
                  MD_n_total += 1
            if re.match("\d+", element):
                  MD_n_total += int(element)      
        #since insertion bases are not included in MD accounting, we need to add them 
        #to figure out the pre-repo softclip length
      I_num = re.findall("(\d+)I", z.cigar)  
      if I_num:
            I_num = sum(map(int, I_num))
      else:
            I_num = 0       
      MD_n_total += I_num      
      
      # if the whole softclip was introduced by bwa, no adjustments are necessary
      bwa_softclip_len = 0            
      if MD_n_total < seqLength: 
            bwa_softclip_len = seqLength - MD_n_total
      softclip_len = int(re.search("(\d+)S", z.cigar).group(1))
      if bwa_softclip_len == softclip_len:
            return z
      
      #count how many mismatches are in softclipped region, subtract from NM
      mismatch_in_softclip = 0
      res = re.search("(\d+)S$", z.cigar)
      if res:  
            MD_split.reverse() #if softclip is at end of read, reverse MD
      n=0      
      for element in MD_split:
            if n >= softclip_len - bwa_softclip_len: 
                  break            
            if element == "0":
                  continue
            if re.match("[ACTG]", element):
                  mismatch_in_softclip += 1
                  n += 1
            if re.match("\d+", element):
                  n += int(element)
            
                       
      NM -= mismatch_in_softclip                 
           
      #subtract dropped_indel_bases from NM
      NM -= z.dropped_indels_bases
      
      #update MD    
      n=0
      while MD_split:  #this loop will never get to all MD elements, breaksouts end it
            element = MD_split[0]
            if element == "0":
                  pass
            if re.match("[ACTG]", element):
                  n += 1
            if re.match("[1-9]+", element):
                  n += int(element)
            
            if n < (softclip_len - bwa_softclip_len): 
                  MD_split.pop(0)
            elif n == (softclip_len - bwa_softclip_len): 
                  MD_split.pop(0)
                  break
            elif n > (softclip_len - bwa_softclip_len):
                  MD_split.pop(0)
                  remain = n - softclip_len + bwa_softclip_len
                  MD_split.insert(0, str(remain))
                  break
      
      #if softlip was at end, reverse back MD                 
      if res:
            MD_split.reverse() 
      
      new_MD = ''.join(MD_split)
     
      #update z object      
      z = samIterator.editTag(z, 'MD', new_MD)
      z = samIterator.editTag(z, 'NM', str(NM))      
      
      # modify XN and XM tags, update z object
      if 'XN' in tag_details:
            XN = tag_details['XN'] 
            XN -= (softclip_len-bwa_softclip_len)
            z = samIterator.editTag(z, 'XN', str(XN))
      if 'XM' in tag_details:
            XM = tag_details['XM']         
            XM -= mismatch_in_softclip
            z = samIterator.editTag(z, 'XM', str(XM))
      
      return z
                 
def samSpecChecks(z1,z2,flagType,checksFile):
      
      #save flags as they were before repositioning
      z1Flag = []
      z2Flag = []
      try:
            z1Flag = flagType[z1.flag]
      except KeyError:
            flagError = z1.qname + ","+ str(z1.flag)+"\n"
            checksFile.write(flagError)
      try:
            z2Flag=flagType[z2.flag]
      except KeyError:
            flagError = z1.qname + ","+ str(z1.flag)+"\n"
            checksFile.write(flagError)

      
      # deals with cases where one read is mapped and the other is not
      if (z1.rname <> "*" and z1.cigar <> '*' and z2.cigar == '*') or (z2.rname <> "*" and z2.cigar <> '*' and z1.cigar == '*'): 
            
            #assign each alignment object to a generic name based on its mapping status
            if z1.cigar == "*":
                  z_unmapped = z1
                  z_mapped = z2
            else:
                  z_unmapped = z2
                  z_mapped = z1
                  
            #assign mapq of mapped read based on its SM tag	
            tagDict = samIterator.getTagDetails(z_mapped)
            if 'SM' in tagDict:
                  if int(z_mapped.mapq) != tagDict['SM']:
                        z_mapped.mapq = tagDict['SM']

            flagCode1=[]
            flagCode2=[]
            
            #adjust flag of mapped read: remove proper pair (2) and strand of mate (32) bits, add mate is unmapped bit (8)
            try:
                  flagCode1 = flagType[z_mapped.flag]
                  if 8 not in flagCode1:
                        z_mapped.flag += 8

                  if 2 in flagCode1:
                        z_mapped.flag -= 2
                  if 32 in flagCode1:
                        z_mapped.flag -= 32
                    
            except KeyError:
                  flagError = z_mapped.qname + ","+ str(z_mapped.flag)+"\n"
                  checksFile.write(flagError)

            #adjust flag of unmapped read: remove proper pair (2) bit, add unmapped bit (4)  
            try:
                  flagCode2 = flagType[z_unmapped.flag]
                  if 4 not in flagCode2:
                        z_unmapped.flag += 4
                  
                  if 2 in flagCode2:
                        z_unmapped.flag -= 2                      
                      
            except KeyError:
                  flagError = z_unmapped.qname + ","+ str(z_unmapped.flag)+"\n"
                  checksFile.write(flagError)

            # replace the strand (16 bit) flag of the unmapped read with that of the mapped read. revcom unmapped read if its strand has changed.   
            if 16 in flagCode1 and 16 not in flagCode2:
                  z_unmapped.flag += 16   
                  if 32 not in flagCode2: # unmapped reads should have rR or nothing
                        z_unmapped.flag +=32                  
                  z_unmapped.qual = revQual(z_unmapped.qual)
                  z_unmapped.seq = revcomp(z_unmapped.seq)                  
            elif 16 not in flagCode1 and 16 in flagCode2:
                  z_unmapped.flag -= 16
                  if 32 in flagCode2: # unmapped reads should have rR or nothing
                        z_unmapped.flag -=32                  
                  z_unmapped.qual = revQual(z_unmapped.qual)
                  z_unmapped.seq = revcomp(z_unmapped.seq)       
                  
            #inheritence: replace the chromosome & position of the unmapped read with those of the mapped read
            z_unmapped.rname = z_mapped.rname
            z_unmapped.pos = z_mapped.pos
            
            #assign mapQ to 0 (unmapped read), MateRefSeqName to "=" (both reads), insert size to 0 (both reads), mpos to position of mapped (both  reads)
            z_unmapped.mapq = 0
            z_unmapped.mrnm = "="
            z_mapped.mrnm = "="
            z_unmapped.mpos = z_mapped.pos
            z_mapped.mpos = z_mapped.pos
            z_unmapped.isize = 0
            z_mapped.isize = 0
                  
            #replace the mapped and unmapped alignment objects to their respective names
            if z1.cigar == "*":
                  z1 = z_unmapped
                  z2 = z_mapped 
            else:
                  z2 = z_unmapped
                  z1 = z_mapped 

      
      #deals with cases where both reads are unmapped (builds flag from scratch)
      elif any([z1.rname == "*", z1.cigar == "*", z1.setUnmapped == 1]) and any([z2.rname == "*", z2.cigar == "*", z2.setUnmapped == 1]):
            
            z1.flag = 1 + 4 + 8 + 64  
            z2.flag = 1 + 4 + 8 + 128

            if 16 in z1Flag:
                  z1.qual = revQual(z1.qual)
                  z1.seq = revcomp(z1.seq)
            if 16 in z2Flag:
                  z2.qual = revQual(z2.qual)
                  z2.seq = revcomp(z2.seq)
                  
            if 512 in z1Flag:
                  z1.flag += 512
            if 512 in z2Flag:
                  z2.flag += 512                  

            pair = [z1,z2]
            for z in pair:    
                  z.cigar = "*"
                  z.pos = '0'
                  z.mapq = '0'
                  z.rname = '*'
                  z.mrnm = '*'
                  z.mpos = '0'
                  z.isize = 0

                  z = clearTags(z)


      # deals with cases where neither read is unmapped S
      elif z1.cigar <> "*" and z2.cigar <> "*": 
      
            # recalculate positioning
            z1.mpos = z2.pos
            z2.mpos = z1.pos
            
            z1.mrnm = z2.rname
            z2.mrnm = z1.rname
            
            if z1.mrnm == z1.rname:
                  z1.mrnm = "="
            if z2.mrnm == z2.rname:
                  z2.mrnm = "="
      
            # recalculate insert size / template length (only for pairs on same chromosome)
            if z1.rname == z2.rname:
                        z1.genomeCoverage = convertCigar.getCigarValues(z1.cigar)
                        z2.genomeCoverage = convertCigar.getCigarValues(z2.cigar)
                        
                        z1_start = z1.pos
                        z2_start = z2.pos
                        
                        flagCode1 = flagType[z1.flag]
                        if 16 in flagCode1:
                              z1_start = int(z1.pos) + z1.genomeCoverage
                              
                        flagCode2 = flagType[z2.flag]
                        if 16 in flagCode2:
                              z2_start = int(z2.pos) + z2.genomeCoverage    
                    
                        z1.isize = z2_start - z1_start
                        z2.isize = z1_start - z2_start
                            
            else:
                  z1.isize=0
                  z2.isize=0  
                  
      else: #it would appear that nothing is leftover from the 3 categories. yay!
            global no_formatting
            print "read %s outside all categories" %z.qname 
      
      #All read pairs go through final part of funtion:

      if z1.flag not in flagType:
            #print "z1.flag not known:",z1.flag, z1.qname, z1Flag
            flagError = z1.qname + ","+ str(z1.flag)+","+ str(z1Flag)+"\n"
            checksFile.write(flagError)
      if z2.flag not in flagType:
            #print "z2.flag not known:",z2.flag, z2.qname, z2Flag
            flagError = z2.qname + ","+ str(z2.flag)+","+ str(z2Flag)+"\n"
            checksFile.write(flagError)

      #set ZJ:Z:Q tag for unmapped reads      
      pair = [z1,z2]
      for z in pair:
            if z.setUnmapped == 1:
                  ZJtag = "ZJ:Z:Q"
                  tagDict1 = samIterator.getTagDetails(z)
                  tags = z.tags
                  tags.append(ZJtag)
                  z.tags = tags
                  z = clearTags(z)

      return z1, z2


def clearTags(z):
      # if a decision is made to remove tag information due to setting a read to unmapped
      # for example, RG tag needs to be maintained if it exists.
      # for GSC custom pipline, ZJ tag also maintained.

      zTags = []

      for t in z.tags:
            i = t.split(':')
            tag = i[0]
            value = i[2]
            if tag == 'RG': 
                  zTags.append(t)
            if tag == 'ZJ' and value == 'Q':
                  zTags.append(t)

      z.tags = zTags

      return z



def checkOrientation(z1, z2, flagCode1,flagCode2):
      # idea: add flagCode lists to z object

      # proper orientation defined as reads aligning as such:
      #    --->     
      #--------------------------
      #             <----
      properOrientation=0
      if (16 in flagCode1 and 32 not in flagCode1) and (32 in flagCode2 and 16 not in flagCode2) and z2.pos < z1.pos:
            properOrientation=1
      if (16 in flagCode2 and 32 not in flagCode2) and (32 in flagCode1 and 16 not in flagCode1) and z1.pos < z2.pos:
            properOrientation=1

      return properOrientation


def removeDups(z,flagCode):

      if 1024 in flagCode: # remove dup flag
            z.flag-=1024

      return z



def properPairCheck(z1, z2, flagType, checksFile):
      # A final check though there might be some redundancy with some aspects of samSpecChecks
      # This is run on ALL reads
      flagCode1=[]
      flagCode2=[]
      max_sep = int(max_separation)
      
      # set flag codes 
      try:
            flagCode1 = flagType[z1.flag]
      except KeyError:
            flagError = z1.qname + ","+ str(z1.flag)+"\n"
            checksFile.write(flagError)
      try:
            flagCode2 = flagType[z2.flag]
      except KeyError:
            flagError = z2.qname + ","+ str(z2.flag)+"\n"
            checksFile.write(flagError)

      # check if both ends are still aligned and adjust parametres as necessary
      if z1.rname <> '*' and z1.cigar <> '*' and z2.rname <> '*' and z2.cigar <> '*':
            tagDict1 = samIterator.getTagDetails(z1) #gets the optional tags at the end of the sam record
            tagDict2 = samIterator.getTagDetails(z2)

            properOrientation = checkOrientation(z1,z2,flagCode1,flagCode2)
            if properOrientation==1 and z1.mrnm == "=" and z2.mrnm == "=" and abs(z1.isize) <= max_sep and abs(z2.isize) <= max_sep:
                  # read has proper orientation
                  if 2 not in flagCode1: 
                        z1.flag += 2
                  if 2 not in flagCode2: 
                        z2.flag += 2
                        # proper pair
                  if 'SM' in tagDict1 and 'SM' in tagDict2:
                        if tagDict1['SM'] > 0 and tagDict2['SM'] > 0:
                              mapq = tagDict1['SM'] + tagDict2['SM']
                              if mapq > 60:
                                    mapq = 60
                              z1.mapq = mapq
                              z2.mapq = mapq
                        else:
                              # at least one read has SM=0
                              # leave mapQ values unchanged
                              pass
            else:
                  # do nothing - keep an eye on this!
                  # code below shouldn't be needed  <-- but it is
                  if z1.mrnm != "=" or abs(z1.isize) > max_sep or properOrientation ==0:
                        if 2 in flagCode1: 
                              z1.flag -= 2
                  if z2.mrnm != "=" or abs(z2.isize) > max_sep or properOrientation ==0:
                        if 2 in flagCode2:
                              z2.flag -= 2
                  # pass


      return z1, z2




def add_ZJ_tag(z):
      # If read was junction aligned add this custom feature requested by Nina Theissen 
      # which lists first and last bases of intron(s) skipped
      ZJtag = convertCigar.get_ZJ_tag(z.tempCigar, z.pos, z.intron_bool)

      tags = z.tags
      tags.append(ZJtag)
      z.tags = tags

      return z


def addToBed(z):
      """
       add to bed file for each read its alignment start and each junction read through start
       for use in IGV for QC purposes.

       Acquires up to three junction coordinate positions

       Load BED file into IGV and use CTL-F and CTL-B to move between coordinates
       """
      multipleBed=[]
      bedPos = z.rname + "_" + str(z.pos)
      cigar = z.cigar

      pos2=int(z.pos)
      pos3=0

      c = convertCigar.read_cigar(cigar)
      c1=[]
      for k in c:
            c1.append([int(k[:-1]),k[-1]])

      #print c1
      posCount=1

      for i in c1:
            if posCount==1:
                  if i[1]=='M' or i[1] == "S":
                        pos2+=int(i[0])
                  if i[1]=="N":
                        pos2+=int(i[0])
                        pos3+=pos2
                        posCount+=1
            else:
                  if i[1]=='M' or i[1] == "S":
                        pos3+=int(i[0])
                  if i[1]=="N":
                        pos3+=int(i[0])

      multipleBed.append(bedPos)

      if pos2 <> int(z.pos):
            bedPos = z.rname + "_" + str(pos2)
            multipleBed.append(bedPos)

      if pos3<>0:
            bedPos = z.rname + "_" + str(pos3)
            multipleBed.append(bedPos)

      #print "==>", multipleBed
      return multipleBed



def makeBed(bedCoord,outputStub):
      f = outputStub + ".repos.bed"
      f1 = open(f,'w')

      header = "track name="+outputStub+ ".repos"

      bedCoord = set(bedCoord)

      for j in bedCoord:
            i = j.split('_')
            chromosome = '_'.join(i[0:-1])
            coordinate = int(i[-1])

            line = ''
            if "chr" in chromosome:
                  line += chromosome 
            else:
                  line += "chr"+chromosome

            line +='\t'+str(coordinate-1)+'\t'+str(coordinate)
            f1.write(line)
            f1.write("\n")


def revcomp(seq):
      # reverse complement sequence (in captitals)
      # i.e. convert: 'TAATTCCCAAGTTCCTTATGGCCAATGGTCAGTTGGTGAAGATGCTGCTT'
      #           to: 'AAGCAGCATCTTCACCAACTGACCATTGGCCATAAGGAACTTGGGAATTA'
      trans={'N':'N','A':'T','T':'A','G':'C','C':'G'}
      j=len(seq)-1
      revcompseq=trans[seq[j:]]
      i=0
      while j > 0:
            j-=1
            i-=1
            revcompseq+=trans[seq[j:i]]
      return revcompseq


def revQual(qual):
      # reverse the quality string
      i=0
      j=len(qual)-1
      revQual=qual[j:]
      while j > 0:
            j-=1
            i-=1
            revQual+=qual[j:i]
      return revQual


def reposition(jObj,junctionStart,junctionCoordinate,z,checksFile):
      """ 
       Put together new position and create new cigar
      """
      tempSeqLength = 0
      origCigar = z.cigar

      #if z.cigar == '8M1I91M':
            #z.cigar = '100M'

      if printInfo==1:
            print " "
            print "Repositioning....",z.qname, z.flag

      noExons = len(jObj.genomeRange)/2
      genomeRange = jObj.genomeRange
      if printInfo==1:
            print "genomeRange->:", genomeRange

      e=0
      exonFlag=0
      exonJunctionSeqLength =0
      start=0

      N = []  # Gap length that read spans; N value in new CIGAR string
      M = []  # Matching sequence; M value in new CIGAR string
      Mlength=0
      while e <= len(genomeRange):#/2+2:
            if printInfo==1:
                  print e, genomeRange
                  print "junctionStart",junctionStart
            if e >= len(genomeRange):
                  break
            eLength = genomeRange[e+1]-genomeRange[e]+1
            exonJunctionSeqLength += eLength

            if printInfo==1:
                  print "exonJunctionSeqLength",exonJunctionSeqLength
                  print "1-->",genomeRange[e],genomeRange[e+1],exonJunctionSeqLength
                  print "2-->",junctionStart, exonJunctionSeqLength, exonFlag

            #deals with the first exon that the read aligns to
            if junctionStart <= exonJunctionSeqLength and exonFlag == 0: 
                  if printInfo==1:
                        print junctionStart,"*"

                  start = int(junctionStart)-exonJunctionSeqLength+genomeRange[e+1] #-> hg19 test
                  #start2 = genomeRange[e] + int(junctionStart) -1 #this would only work if the read starts in the first exon of the juntion sequence

                  if start + (seqLength-z.readAlignedToPadStart) > genomeRange[e+1]: #the read extend beyond this exon
                        Mlength = genomeRange[e+1] - start + 1                      
                  elif start + (seqLength-z.readAlignedToPadStart) <= genomeRange[e+1]: #the read starts and ends in a single exon      
                        Mlength = seqLength - z.readAlignedToPadStart
                              
                  if printInfo==1:
                        print "Mlength",Mlength

                  M.append(Mlength)
                  tempSeqLength+=Mlength + z.readAlignedToPadStart

                  exonFlag=1
            
            #deals with all subsequent introns and exons that the read aligns to
            elif exonFlag==1 and tempSeqLength < seqLength:

                  # We seem have now entered junction N space.  
                  # Please return to your seats with your seat belt on and bags stowed
                  
                  junctionGapSize=genomeRange[e] - genomeRange[e-1] -1 
                  N.append(junctionGapSize)

                  if seqLength - tempSeqLength > eLength: #the read extends into the exon after (e+1)/2
                        if printInfo==1:
                              print "==>",tempSeqLength, eLength + z.readAlignedToPad
                              print "eLength,z.readAlignedToPad:",eLength,z.readAlignedToPad
                        M.append(eLength)  #+ z.readAlignedToPad) 
                        tempSeqLength+=eLength
                  else:                                    #the read ends in this the e+1th (current) exon
                        M.append(seqLength-tempSeqLength)
                        tempSeqLength+=eLength
   
            e+=2

      z.softclip = z.readAlignedToPadStart

      if z.readAlignedToPad > 0 and Mlength < seqLength:
            #M.append(padOverhang)      this is the bug which produces double M's instead of end softclips
            z.softclipEnd = z.readAlignedToPad
      
      #this is to get rid f 0N's and merge adjacent M's (this is caused by introns of size 0 in reference junction objects)  
      while 0 in N:
            ind = N.index(0)
            if ind ==0:
                  M_ind = 0
            else:
                  M_ind = ind
            del N[ind] 
            newM = M[M_ind] + M[M_ind+1]
            M[M_ind] = newM
            del M[M_ind+1] 
            
            if printCigarAndFlagError:
                  zeroN_error_line = "INFO read aligned to junction sequence with 0N\t%s\n" %z.qname
                  checksFile.write(zeroN_error_line)

      tempReadlength = seqLength

      if start == 0:
            z.cigar = "*"
            z.pos = '0'
            z.mapq = '0'
            z.setUnmapped = 1
            return z

      #put new locations into cigar
      n=0
      cigar=""
      d=0
      if z.softclip>0: 
            cigar += str(z.softclip)+'S'
      mCount=0
      for c in M:

            c1=''
            if mCount==0 and len(M) > 1:

                  #c1 = str(c-z.softclip)+"M"
                  c1 = str(c)+"M"

                  #print "!!!",c1
                  mCount+=1
            elif mCount==1 and len(M) > 1:
                  #c1 = str(c+z.softclip)+"M"

                  c1 = str(c)+"M"
                  mCount+=1
            else:
                  c1 = str(c)+"M"

            cigar+=c1
            d+=c
            if d == seqLength:
                  break
            if len(N)>n:
                  n1 = str(N[n])+"N"
                  cigar+=n1
                  n+=1

      if printInfo==1:
            print cigar, z.cigar

      """ put a check here on [S + M + I] total == readlength """
      if z.softclipEnd > 0 and sum(M) < seqLength:  
            cigar += str(z.softclipEnd)+'S'
      
      #save temp cigar for calculating ZJ tag later
      z.tempCigar = cigar
      
      #merge gaps in temp cigar with I's D's and S's from original cigar
      if z.cigar <> str(seqLength)+"M" and z.cigar <> "*" and z.cigar <> "" and cigar <> z.cigar:
            if printInfo==1:
                  print "Converting cigar...", cigar, z.cigar

            #if origCigar == '8M1I91M':
            #    cigar = '67M1116N33M'

            (z.cigar, z.dropped_indels_num, z.dropped_indels_bases, z.intron_bool, z.add_to_pos) = convertCigar.convert_new(cigar,z.cigar)	
      else:
            z.cigar = cigar

      if z.cigar == '*' or cigarCheck(z.cigar)==0:
            #Need to check these; something went wrong in the cigar formatting
            if printCigarAndFlagError:
                  cigarLine = 'a', z.qname, str(z.flag), z.rname, z.cigar, cigar, origCigar
                  cigarLine = "\t".join(cigarLine)
                  cigarFile.write(cigarLine)
                  cigarFile.write('\n')   
            z.cigar = "*"
            z.pos = '0'
            z.mapq = '0'
            z.rname = '*'
            z.mrnm = '*'
            z.setUnmapped = 1
            return z    	
   #   start = start + z.add_to_pos
      z.pos = start  #+z.softclip

      return z


def cigarCheck(cigar):
      # identify bad cigars
      c = convertCigar.read_cigar(cigar)
      length=0

      for k in c:
            if k[-1]=='I' or k[-1]=='M' or k[-1]=='S':
                  length += int(k[:-1])

      if length <> seqLength:
            return 0
      elif "-" in cigar:
            return 0
      else:
            return 1


def createHeader(chrOrder,f_header,SQ_info,sam_file,output_dir,index_folder,max_sep):
      sortedChromosomes = chrOrder
      sam_file_name = os.path.basename(sam_file)
      index_file_folder = os.path.basename(index_folder)
      index_file_folder = os.path.basename(index_folder)+"/"+index_file_folder
      PG_count=0
      SQ_count=0
      PG_line = "@PG\tID:1\tPN:JAGuaR\tVN:"+version+"\tCL:convertJunction.py -f "+sam_file_name+" -i "+index_file_folder+" -m "+max_sep+"\n"

      for h in f_header:
            # header line order @HD, @SQ, @RG, @PG, @CO.
            if '@HD' in h:
                  newSam.write(h)
            elif '@SQ' in h:
                  if SQ_count==0:
                        for s in SQ_info:
                              newSam.write(s)
                        SQ_count += 1
                  else:
                        continue
            elif '@RG' in h:
                  newSam.write(h)
            elif '@PG' in h:
                  if PG_count==0:
                        newSam.write(PG_line)
                        PG_count = PG_count+1
                  if PG_count > 0:
                        newSam.write(h)
            else:
                  newSam.write(h)  


def sam2bam(sam_files):
      file_list=sam_files.split(",")
      for sam_file in file_list:
            new_sam_file = sam_file[:-4]+".sam"
            bamfile = sam_file[:-4]+".bam"
            sortedbam = sam_file[:-4]+".sorted"
            print " "
            print "==> Converting " + new_sam_file + " to sorted and indexed BAM..."
            command = sam_exec+" view -bT "+ RefDefault +" "+ new_sam_file + " -o " +  bamfile
            print command
            print " "
            os.system(command)
            command = sam_exec+" sort "+bamfile+ " "+sortedbam
            print command
            print " "
            os.system(command)
            command = sam_exec+" index "+sortedbam+".bam"
            print command
            print " "
            os.system(command)


def convertSam(chrOrder,SQ_info,max_sep,index_folder,sam_file,new_sam_file,lineCountMax,readList,chrTranscripts,jidx,s,bedFile,checksFile):
      """
       Go through sam file
       Read two lines at a time
       Works for PET only and reads must be sorted so pairs are one after each other by line
      """
      timeCount=linesToReport
      chrAlign=0
      global seqLength

      # create log file
      #logFile = "log.txt"
      #global logFile
      #logFile.write("{0}\n".format) # print various output for QC/debugging
      #outputLog = open(logFile,'w')

      flagDesc,flagType = samFlags.flags()

      # check if the input is in SAM or BAM format based on file extension
      f=''
      f_header = ''
      fileType = sam_file[-3:]
      if fileType == "bam":
            proc = subprocess.Popen(               
                  [sam_exec, 'view', '-h', sam_file],               
                  shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)  
            f = proc.stdout
            proc_h = subprocess.Popen(
                  [sam_exec, 'view', '-H', sam_file],
                  shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
            f_header = proc_h.stdout
      elif fileType == "sam":
            f = open(sam_file,'r')
            try:
                  proc_h = subprocess.Popen(
                        [sam_exec, 'view', '-SH', sam_file],
                        shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
            except OSError, e:
                  print "Error: %s" % e
                  print e.errno, e.strerror
                  sys.exit()
            f_header = proc_h.stdout
      else:
            print "Cannot recognize file format.  File extension must be either .sam or .bam"

      print "Reading "+fileType+" file and repositioning..."

      global newSam 
      newSam = open(new_sam_file,'w')
      output_dir = os.path.dirname(new_sam_file)
      print "Creating header for the repositioned sam..."
      createHeader(chrOrder,f_header,SQ_info,sam_file,output_dir,index_folder,max_sep)

      new_sam_jo_file = new_sam_file + "jo"

      bedCoord = []
      bedInfo = {}
      lineCount=0

      z1=align() #create an empty object
      z2=align() #create an empty object

      linePair=0
      line1=''
      aa = 0
      bb=0      
      for line in f:

            if line[0]=="@":
                  # if printing the header, need to recalculate the length (LN) heading
                  continue

            if lineCount == timeCount:
                  timeCount+=linesToReport
                  progressUpdate(lineCount,timeCount)

            if lineCount==int(lineCountMax):
                  #newSam.write('\n')
                  break

            if linePair==0:
                  line1 = line
                  linePair=1
                  continue

            lineCount+=2	

            x = line1.split()
            y = line.split()
            s.totalReads+=2

            samIterator.readSAM(z1,x, options.mt, options.chrom) #z1 will be modified and will contain the contents of x when this function returns
            samIterator.readSAM(z2,y, options.mt, options.chrom) #z2 will be modified and will contain the contents of y when this function returns

            if lineCount==2:
                  seqLength = int(len(z1.seq))	

            if z1.qname[-2] == "/":
                  z1.qname = z1.qname[:-2]
            if z2.qname[-2] == "/":
                  z2.qname = z2.qname[:-2]

            if len(readList) > 0:
                  if z1.qname in readList:
                        logFile.write("{0}\n".format(line))
                  else:
                        continue

            if z1.qname <> z2.qname:
                  print z1.qname, z2.qname
                  print "Fatal error - no pairs"
                  exit(0)

            pair =[z1,z2]
            for z in pair:

                  # remove dups flag as the identification of duplication should be done post repositioning
                  try:
                        flagCode = flagType[z.flag]
                        z = removeDups(z,flagCode) 
                  except KeyError:
                        flagError = z2.qname + ","+ str(z2.flag)+"\n"
                        checksFile.write(flagError)
                  
                  if z.rname in genomeInfo:
                        chrLength = int(genomeInfo[z.rname])
                        if int(z.pos) > chrLength:  
                       
                              """
		        READ ALIGNED TO JUNCTION REGION OF CHROMOSOME
		        Flag it as such for further downstream processing
		        """

                              #print z.qname, z.flag, z.cigar
                              z.junction = 1
                              s.junctionReads += 1

                              if z.cigar <> "*":
                                    if printInfo==1:
                                          print "----------------------------------------------------"
                                          print "Read aligned to junction reference, reformatting..."
                                          print z.qname, z.flag, z.cigar

                                    #comment1 = "----------------------------------------------------"
                                    #comment2 = "Read aligned to junction reference, reformatting..."
                                    #logFile.write("{0}\n".format(.write("{0}\n".format(comment1))
                                    #logFile.write("{0}\n".format(.write("{0}\n".format(comment2))
                                    #logFile.write("{0}\n".format(.write("{0}, {1}, {2}\n".format(z.qname, z.flag, z.cigar))

                                    s.junctionAligned+=1

                                    z = reformat(z,jidx,checksFile)
                                    
                                    if z.cigar == "*":
                                          z.setUnmapped=1  
                                    else:
                                          z = add_ZJ_tag(z)
                                    
                                    z.pos = int(z.pos) + z.add_to_pos  #position is updated after ZJ is calculated
                                           
                                    #else: #this includes reads in the form "100M", as well as reads with S's and I's and D's, but no N's
                                        #  s.junctionReadNoN +=1

                                    
                              else:
                                    # position given but no cigar - read originally unaligned
                                    s.unMapped+=1


                        elif (int(z.pos)+seqLength) > chrLength:
                              
                              """
                        START OF READ ALIGNED TO GENOME BUT END ALIGNED TO JUNCTION REGION
                        Flag it as such for further downstream processing
                        """
                              # can work on this a little further to capture reads that could be accurately aligned to the end of the chromosome 
                              z.cigar = "*"
                              z.mapq = '0'
                              z.setUnmapped = 1
                              
            z1 = pair[0]
            z2 = pair[1]

            # update positions for all read pairs
            z1.mpos = z2.pos
            z2.mpos = z1.pos
            
            #take care of the flags. positions and other sam specs for read pairs that have undergone repositioning
            juncReadCount = z1.junction + z2.junction
            setUnmapped = z1.setUnmapped + z2.setUnmapped
            if juncReadCount > 0 or setUnmapped > 0:
                  z1, z2 = samSpecChecks(z1, z2, flagType, checksFile) 
                  

            # test all reads for proper pair bit (2)
            z1, z2 = properPairCheck(z1, z2, flagType, checksFile) 

            # update NM and MD tags in reads that have been softclipped by Jaguar
            pair = [z1,z2]
            for z in pair:  
                  if 'S' in z.cigar and z.junction == 1:
                        z = samTagUpdate(z)
            
            for z in pair:	  

                  if bedFile==1:
                        tempCigar = str(seqLength)+"M"
                        if z.cigar <> tempCigar and z.cigar <> "*" and z.junction==1:
                              bedPos = addToBed(z)
                              bedCoord+=bedPos

                  newLine = samIterator.createline(z)
                  newSam.write(newLine)
                  newSam.write('\n')

                  tagDict = samIterator.getTagDetails(z)
                  if 'ZJ' in tagDict:
                        tag = tagDict['ZJ']
                        if tag == 'Q':
                              s.setUnmapped+=1
                        else:
                              # a junction aligned read
                              s.junctionRepositionedReads+=1

                              k = tag.split(",")
                              noJunctions = len(k)
                              if noJunctions in s.junctionNumberCount:
                                    s.junctionNumberCount[noJunctions]+=1
                              else:
                                    s.junctionNumberCount[noJunctions]=1
                              s.junctionCoverageTotal+=noJunctions
                              for kj in k:
                                    if kj in s.junctionUniqueCoverageCount:
                                          s.junctionUniqueCoverageCount[kj]+=1
                                    else:
                                          s.junctionUniqueCoverageCountTotal+=1
                                          s.junctionUniqueCoverageCount[kj]=1

            linePair=0

      print "Complete!"
      print "Total number of reads analyzed: ",s.totalReads

      newSam.close()

      if bedFile==1:
            makeBed(bedCoord,outputStub)

      return s



if __name__=='__main__':
      l1 = "\033[1mJAGuaR - Junction Alignment to Genome for Repositioning RNA-seq Reads\033[0m"  
      l2 = "\033[1mVersion: "+str(version)+"\033[0m"
      print " "
      print l1
      print l2
      print "Date: "+versionDate
      print "Contact: Yaron Butterfield <ybutterf@bcgsc.ca>"
      print info
      print " "
      print "Input: SAM or BAM file of paired end tag (PET) reads aligned to Jaguar reference/junctions with BWA (sampe -s) and sorted by name"
      print "Output: SAM file with genomic coordinates with appropriate reads split by exon"
      print " "
      parser = OptionParser(description=__doc__)
      parser.add_option("-f",dest="f",help="PET sorted .sam or .bam (f)ile. REQUIRED",metavar="file")
      parser.add_option("-o",dest="o",help="(O)utput stub for repositioned file incl. directory, otherwise output in cwd; i.e., [Output tag].repos.sam.  REQUIRED",metavar="o")
      parser.add_option("-i",dest="i",help="(I)ndex folder. REQUIRED",metavar="index")
      parser.add_option("-m",dest="m",help="(M)ax separation. (M)ax separation recommendations:34622 is used for hg19a, 23545 is used for hg18 REQUIRED",metavar="max_separation")

      parser.add_option("-g",dest="g",help="(G)enome reference fasta file with path. Required if using option -b",metavar="genome")
      parser.add_option("--samtools", dest="sam_exec", default='samtools',help="SAMTOOLS executable")

      parser.add_option("-b",dest="b",action="store_true",help="create (B)am after sam conversion",metavar="b")  

      parser.add_option("-c",dest="c",help="chromosome index check, default=1",metavar="check")
      parser.add_option("-d",dest="d",action="store_true",help="print debugging messages to screen",metavar="d")
      parser.add_option("-s",dest="s",action="store_true",help="create cigar.txt (CIGAR errors) and .checks (FLAG errors) files")
      parser.add_option("-x",dest="x",help="filter by comma delimited list of chromosomes, default all chromosomes listed in genome length file",metavar="chromosome")
      parser.add_option("-l",dest="l",help="Number (L)ines/reads to reposition from beginning of -f file",metavar="l")
      parser.add_option("-r",dest="r",help="pull out read(s) and create reads.sam",metavar="r")

      parser.add_option("-n",dest="n",action="store_true",help="NO to repositioning, assumes a repositioned file already exists",metavar="n")
      parser.add_option("-e",dest="e",action="store_true",help="Create bed file of exon-exon junction covered read coordinates",metavar="e")
      parser.add_option("--do_not_check_mito_naming",dest="mt",action="store_true",help="if not specified, mitochondrial chromosome M (chrM) will be renamed MT (chrMT)")
      parser.add_option("--do_not_drop_chr_prefix",dest="chrom",action="store_true",help="If not specified, the 'c-h-r' prefix of any chromosome name will be dropped (chr8 becomes 8)")
      (options, args)=parser.parse_args()

      reads = options.r
      outputStub = options.o
      sam_file= options.f
      indexFolder = options.i 
      lineCount=options.l
      noRepos=options.n
      sam_exec = options.sam_exec
      global printCigarAndFlagError
      printCigarAndFlagError=options.s

      #print options
      #print args
      #print sys.argv
      #subprocess.check_call([sam_exec])
      if options.f is not None and reads <> None:
            pullOutReadFromSam(sam_file, reads)
            sys.exit(-1)
      elif options.f is None or options.i is None:
            print "Missing required information"
            print "Please use -h for command line options"
            sys.exit(-1)

      global RefDefault
      RefDefault=options.g

      global max_separation
      max_separation = options.m

      # do a check here to make sure this file exists
      samExist = os.path.isfile(sam_file)
      if samExist is False:
            print sam_file, "does not exist!"
            sys.exit(-1)

      exonsFastaFile = indexFolder + "/exons.fa"
      print "Reference index folder", indexFolder

      new_sam_file = outputStub+".repos.sam"
      checksFile = open((outputStub+".checks.txt"),'w')

      # get and store some basic statistics and repositioning information
      s = Stats()
      s.referenceIndexFolder = indexFolder
      
      localAlign = 0
      createBam = 0
      bedFile = 0

      if options.b:
            createBam=1
      if options.d:
            printInfo=1
      if options.e:
            bedFile=1

      if reads <> None:
            pullOutReadFromSam(sam_file, reads)
            exit(0)

      if lineCount is None:
            lineCount=999999999999999999

      # get from File?
      # i.e
      #readList = ["SOLEXA12_65:2:1:4996:1305"]
      readList = []


      startTime = time.ctime()
      """
    ----------   The following covers the main component of this code  ----------------
    """
      if noRepos is not True:
            print "Preparing to convert: ",sam_file
            print "Method: ",
            print "Getting junction info (index and sequence) from:"
            print "   ",indexFolder
            chromosomeSizeFile = indexFolder + "/genomeLength.txt"
            sqFile = indexFolder + "/sqHeader.txt"

            if printCigarAndFlagError:
                  cigarFile = open('cigar.txt',"w", 1)

            genomeInfo,chrOrder = getChromosomeInfo(chromosomeSizeFile)
            SQ_info = get_SQ_info(sqFile)
            #print "Getting genome reference sequence"
            #genomeSeq = getRefGenome(RefDefault)

            chrTranscripts = ''

            print "Getting junction information"
            jidx = getJunctionInfo(genomeInfo,indexFolder)

            t1 = time.time()-t0 
            print t1, "seconds"
            print " "

            t0 = time.time()

            s = convertSam(chrOrder,SQ_info,max_separation,indexFolder,sam_file,new_sam_file,lineCount,readList,chrTranscripts,jidx,s,bedFile,checksFile)

            t2 = time.time()-t0
            print t2, "seconds"
            print str(t2/60.0), "minutes"

            s.sam_file = sam_file
            s.new_sam_file = new_sam_file
            s.reposMinutes = str(t2/60.0)
            s.startTime = startTime
            printStats(outputStub, s)

      """
     Samtools specific options

    """
      bam_file = outputStub+".repos.sorted.bam"

      if createBam==1:
            outputFiles = new_sam_file
            sam2bam(outputFiles)

      if printCigarAndFlagError is not True:
            #delete .checks file
            command = "rm "+outputStub+".checks.txt"
            os.system(command)
            