#!/gsc/software/linux-x86_64-centos5/python-science-2.7.1/bin/python

description = """ This function makes sure that the cigar and ZJ tag from a repositioned bam file are in agreement.
This would apply to bam files produced with JAGuaR v2.0.2 and earlier where junctions covered by softclips were not removed fromt he ZJ tag. 
Requires that pysam is installed. For example /gsc/software/linux-x86_64-centos5/python-science-2.7.1/bin/python

Author: Maayan Kreitzman
Creater: June 2013"""
usage =  """"python fixZJtag.py -f bamfile [-s outsuffix] [-o outDir]"""



import subprocess
import os
import sys 
from optparse import OptionParser
import re
import pysam
import time
	
def main():
	timestampBeg = time.strftime('%Y %m %d-%H:%M:%S')
	
	#parse options and initialize
	parser = OptionParser(description=description, usage=usage)
	parser.add_option("-f",dest="f",help=".bam (f)ile. REQUIRED",metavar="file")
	parser.add_option("-o",dest="o",help="output directory including trailing slash. default is the same directory as the input bam",metavar="dir")	
	parser.add_option("-s",dest="s",help="suffix added to bam file name after processing, default: 'ZJfix'",metavar="string", default="ZJfix")

	
	(options, args)=parser.parse_args()
	
	inbam = options.f
	fileType = inbam[-3:]
	
	if fileType == "bam":
		pass
	else:
		print "input needs to be a bam file"
		sys.exit(-1)
		
	out_suffix = options.s
	
	insam = pysam.Samfile(inbam, "rb")
	
	if options.o:
		res = re.search('.*\/(.*?)$', inbam).group(1)
		new_bam_file = re.sub("bam", out_suffix + ".bam", res)
		new_bam_file = options.o+new_bam_file
		logfilename = re.sub("bam", out_suffix + ".log", res)	
		logfilename = options.o+logfilename
		
	else:
		new_bam_file = inbam[:-3]+ out_suffix + ".bam"
		logfilename = inbam[:-3]+ out_suffix +".log"
	
	logfile = open(logfilename, 'w')		
	logfile.write('PROCESSING STARTED: %s\n\n' %timestampBeg)	
	
	newSam = pysam.Samfile(new_bam_file, "wb", template=insam)
		

	#iterate through lines 
	reads_corrected = 0
	totalReads = 0
	for read_obj in insam:
		
		totalReads +=1
		
		#get cigar string and ZJ tag from the read object
		try:
			ZJ = read_obj.opt('ZJ') #if there's no JZ tag, just write the record unchanged
		except KeyError:
			newSam.write(read_obj)
			continue
		
		cigar = read_obj.cigarstring
		cigar = fix_cigar(cigar) #this is due to a bug in pysam that puts the letters before the numbers in the cigar
		
		num_introns = len(re.findall("N", cigar))
		pairs = ZJ.split(",")
		num_ZJ_pairs = len(pairs)
		
		#if number of introns in cigar and ZJ disagree, fix ZJ, otherwise, write the record unchanged. 
		if cigar and num_introns!= num_ZJ_pairs:
			
			read_obj = fix_ZJ(read_obj, cigar, ZJ, logfile) 
			reads_corrected += 1
			
		if read_obj == 0: #this shouldn't happen: a situation where there's a discrepancy without softclipping
			continue
		else:
			newSam.write(read_obj)
		
	timestampEnd = time.strftime('%Y %m %d-%H:%M:%S')
	logfile.write('\nPROCESSING ENDED: %s\n' %timestampEnd)	
	logfile.write("TOTAL READS PROCESSED:\t%s\nTOTAL READS CORRECTED:\t%s\n" %(str(totalReads), str(reads_corrected)))
	
	print "reads corrected: "+str(reads_corrected)
	
def fix_ZJ(read_obj, cigar, ZJ, logfile):
	
	num_introns = len(re.findall("N", cigar))
	pairs = ZJ.split(",")
	num_ZJ_pairs = len(pairs)	 
	C = read_cigar(cigar)
	pos = read_obj.pos + 1 # PYSAM CONVERTS THE POS COORDINATES TO 0-BASED, ZJ COORDINATES READ AS STRINGS AND STAY 1_BASED
	
	introns_dropped = num_ZJ_pairs - num_introns
	
	if re.search("S$", cigar):
		pairs = pairs[:-introns_dropped]
		new_ZJ = ",".join(pairs)
		logfile.write("LINE REPAIRED due to END softclipping: ")
	
	elif re.search("^\d+S", cigar): # there are cases where a cigar with start softclipping should have the intron removed from the end, and others where it should be removed froom the start.		
		
		first_exon_aligned = get_first_intron_start(C)
		if pos + first_exon_aligned  == int((pairs[0].split("-"))[0]): # if the first exon ends in the first intron coordinate, discrepancy is due to insertions and ZJ coordinate pairs should be removed from the end.
			pairs = pairs[:-introns_dropped]	
			new_ZJ = ",".join(pairs)
			logfile.write("LINE REPAIRED due to INSERION: ")
		else:
			pairs = pairs[introns_dropped:] # discrapnacy is actually due to start softclipping
			new_ZJ = ",".join(pairs)
			logfile.write("LINE REPAIRED due to START softclipping: ")
		
	else:  # a cigar and ZJ tag have different number of introns but no softclipping - should have the intron(s) removed from the end.
		pairs = pairs[:-introns_dropped]
		new_ZJ = ",".join(pairs)
		logfile.write("LINE REPAIRED due to INSERION: ")
	
	if not pairs:
		new_ZJ = "ZJ:Z:G"  #this happens if the only junction in the read was removed due to softclipping or insertion
		
	log_line = "qname: %s pos: %s cigar: %s old ZJ: %s new ZJ: %s\n" %(read_obj.qname, str(pos), cigar, ZJ, new_ZJ) # writes line before fix, followed by new ZJ
	logfile.write(log_line)	
	#print log_line
	
	read_obj.tags = read_obj.tags[:-1] #get rid of old ZJ tag from read object tags array
	read_obj.tags = read_obj.tags + [("ZJ",new_ZJ)] # add new ZJ tag to the read object tags array 			
		
	return read_obj

def read_cigar(cigar):
        splitlist = re.split('([MIDNSHP])', cigar)
        newlist = [splitlist[x]+splitlist[x+1] for x in range(0, len(splitlist)-1, 2)]
        return newlist

def fix_cigar(cigar):

        splitlist = re.split('([MIDNSHP])', cigar)
        new = ''
        saved_letter = ''
        for elem in splitlist:
                if re.search('\d+', elem):
                        new += elem
                        new += saved_letter
                elif re.search('[MIDNSHP]', elem):
                        saved_letter = elem
        return new

def get_first_intron_start(C):
	
	exonAligned = 0
	for element in C:
	       
	        if element[-1]=='M':
			exonAligned += int(element[:-1])
			
		elif element[-1]=='S':
			pass
		
		elif element[-1]=='I':
			pass
		
		elif element[-1]=='D':
			exonAligned += int(element[:-1])
		
		elif element[-1]=='N':
			return exonAligned		             
		
if __name__ == "__main__":
	main()